#include <iostream>
#include <iomanip>
#include <cmath>
#include <unistd.h>
#include <string.h>
#include <set>
#include <pqxx/pqxx>
#include <time.h>
#include <bitset>
#include <sys/stat.h>
#include <ftw.h>
#include <stdlib.h>
#include <time.h>
#include "fitsio.h"
#include "healpix_map.h"

using namespace std;
using namespace pqxx;

string binary( unsigned long n );
int cfind(string str, string sub);
int insertOne(string filepath, string filename, long hid, double ra1, double ra2, double dec1, double dec2, string date);
int insertOne(string filepath, string filename, long hid, double ra, double dec, double rotation, string time);
int query_engine(string sources[], double center_RA, double center_DEC, double X_size, double Y_size, string begin_Time, string end_Time);
int query_engine(string source, int hid, double center_RA, double center_DEC, double RA_size, double Dec_size, string begin_Time, string end_Time);
int standard_query_engine(string source, int hid, double center_RA, double center_DEC, double RA_size, double Dec_size, string begin_Time, string end_Time);
int grammarAnalysis();
bool haveCommonEdge(const unsigned long hid1, const unsigned long hid2);
set<unsigned long> calculateCoveredPixels(pointing pt [], const double &rotation);
void findCoveredByEdge(const pointing &pt1, const pointing &pt2, set<unsigned long> &s);
void findCoveredByCenter(const pointing &center, const pointing vertexes[], set<unsigned long> &s, const double &rotation);
void standardToRotatedCoordiantes(const double &x1, const double &y1, double &x2, double &y2, const double &rotation);
void rotatedToStandardCoordiantes(const double &x1, const double &y1, double &x2, double &y2, const double &rotation);
int selectRows(const result &R, double center_RA, double center_DEC, double RA_size, double Dec_size);

//static int n = 0;
static const double x_length = 12000/3600;
static const double y_length = 5300/3600;
//static const double x_length = 0.125;
//static const double y_length = 0.0625;
static const double PI =  3.14159265359;
static const double PRECISION = 0.0001;
static  int NSIDE = 8;
static const Healpix_Map<unsigned long> HB(NSIDE, NEST);
static  string DB_TABLE = "fileindex_500g_1x0d5_8";
static const double PERCENTAGE = sqrt(0.25);

/**Algorithm 1 */
set<unsigned long> calculateCoveredPixels(pointing pt[], const double & rotation)
{
    set<unsigned long> s;
    pointing center(0,0);
    for (int i = 0; i < 4; i++) {
        findCoveredByEdge(pt[i], pt[i+1], s);
        center.phi += pt[i].phi;
        center.theta += pt[i].theta;
    }
    center.phi /= 4;
    center.theta /= 4;
    findCoveredByCenter(center, pt, s, rotation);
    return s;
}
void findCoveredByEdge(const pointing &pt1, const pointing &pt2, set<unsigned long> &s)
{
    static unsigned long hid1, hid2;
    hid1 = HB.ang2pix(pt1);
    hid2 = HB.ang2pix(pt2);
    s.insert(hid1);
    s.insert(hid2);
    static int i = 0;
    if((abs(pt1.theta-pt2.theta) > PRECISION || abs(pt1.phi-pt2.phi) > PRECISION)
        && hid1 != hid2 && !haveCommonEdge(hid1, hid2) ) {
        pointing middlePoint;
        middlePoint.phi = (pt1.phi + pt2.phi) / 2;
        middlePoint.theta = (pt1.theta + pt2.theta) / 2;
        findCoveredByEdge(pt1, middlePoint, s);
        findCoveredByEdge(middlePoint, pt2, s);
    }
}
void findCoveredByCenter(const pointing &center, const pointing vertexes[], set<unsigned long> &s, const double &rotation)
{
    unsigned long hidCen;
    unsigned long hidVer[4];
    hidCen = HB.ang2pix(center);
    s.insert(hidCen);
    for (int i = 0; i < 4; i++) {
        hidVer[i] = HB.ang2pix(vertexes[i]);
        s.insert(hidVer[i]);
    }
    static int k = 0;
    for (int i = 0; i < 4; i++) {
        if ( (abs(center.theta-vertexes[i].theta) > PRECISION ||
              abs(center.phi-vertexes[i].phi) > PRECISION)
            && hidCen != hidVer[i] && !haveCommonEdge(hidCen, hidVer[i])) {
            pointing centerPoint, vertexPoint1, vertexPoint2;
            centerPoint.phi = (center.phi + vertexes[i].phi) / 2;
            centerPoint.theta = (center.theta + vertexes[i].theta) / 2;
            double center_ra = centerPoint.phi / PI * 180;
            double center_dec = 90 - centerPoint.theta / PI * 180;
            double v_ra = vertexes[i].phi / PI * 180;
            double v_dec = 90 - vertexes[i].theta / PI * 180;
            double temp_dec, temp_ra, temp_vdec, temp_vra;
            double nv_dec1, nv_dec2, nv_ra1, nv_ra2;
            standardToRotatedCoordiantes(center_dec, center_ra, temp_dec, temp_ra, rotation);
            standardToRotatedCoordiantes(v_dec, v_ra, temp_vdec, temp_vra, rotation);
            rotatedToStandardCoordiantes(temp_dec, temp_vra, nv_dec1, nv_ra1, rotation);
            rotatedToStandardCoordiantes(temp_vdec, temp_ra, nv_dec2, nv_ra2, rotation);

            vertexPoint1.phi = nv_ra1 / 180 * PI;
            vertexPoint1.theta = (90-nv_dec1) * PI /180;
            vertexPoint2.phi = nv_ra2 / 180 * PI;
            vertexPoint2.theta = (90-nv_dec2) * PI /180;
            pointing vertexTemp[] = {center, vertexPoint1, vertexes[i], vertexPoint2};
            findCoveredByCenter(centerPoint, vertexTemp, s, rotation);
        }
    }
}

/* Algorithm 2: Index building fuction*/
int  fn(const char *arg, const struct stat *sb, int flag)
{
    if(flag == FTW_D)
        cout << arg << " ---- directory" << endl;
    else if (flag == FTW_F)
    {
        char *file = new char [500];
        strcpy(file, arg);
        if (string(file).find(".fits.fz") == string::npos)
            return 0;
        fitsfile *fptr;
        int status = 0;  /* MUST initialize status */
        double ra, dec;
        char obstime[100];
	//read FITS header
        fits_open_file(&fptr, strcat(file,"[COMPRESSED_IMAGE]"), READONLY, &status);
        fits_read_key(fptr, TDOUBLE, "ra_targ", &ra, NULL, &status);
        fits_read_key(fptr, TDOUBLE, "dec_targ", &dec, NULL, &status);
        fits_read_key(fptr, TDOUBLE, "rotation", &rotation, NULL, &status);
        fits_read_key(fptr, TSTRING, "date-obs", obstime, NULL, &status);   

        double ra1, ra2, ra3, ra4, dec1, dec2, dec3, dec4, ra_temp, dec_temp;
        int n = 0;
        standardToRotatedCoordiantes(dec, ra, dec_temp, ra_temp, rotation);
        rotatedToStandardCoordiantes(dec_temp - x_length/2, ra_temp - y_length/2, dec1, ra1, rotation);
        rotatedToStandardCoordiantes(dec_temp - x_length/2, ra_temp + y_length/2, dec2, ra2, rotation);
        rotatedToStandardCoordiantes(dec_temp + x_length/2, ra_temp - y_length/2, dec3, ra3, rotation);
        rotatedToStandardCoordiantes(dec_temp + x_length/2, ra_temp + y_length/2, dec4, ra4, rotation);
  
        pointing pt1;
        pt1.phi = ra1* PI / 180;
        pt1.theta = (90-dec1) * PI /180;
        pointing pt2;
        pt2.phi = ra2 * PI / 180;
        pt2.theta = (90-dec1) * PI / 180;
        pointing pt3;
        pt3.phi = ra1 * PI / 180;
        pt3.theta = (90-dec2) * PI / 180;
        pointing pt4;
        pt4.phi = ra2 * PI / 180;
        pt4.theta = (90-dec2) * PI / 180;
        pointing pt[] = {pt1, pt2, pt3, pt4, pt1};
        static int i = 0;
        int j = 0;
        set<unsigned long> hids = calculateCoveredPixels(pt, rotation);
        string fileStr(arg);

        for (std::set<unsigned long>::iterator it=hids.begin(); it!=hids.end(); ++it) {
            insertOne(fileStr, "", *it, ra, dec, rotation, obstime);
        delete []file;
        fits_close_file(fptr, &status);
        if (status)          /* print any error messages */
            fits_report_error(stderr, status);
        return(status);
    }
    return 0;
}


/* Algorithm 3: querying function*/
int query_engine(string sources[], double center_RA, double center_DEC, double X_size, double Y_size, string begin_Time, string end_Time)
{
    cout << "[DEBUG]  "  << '\t' << center_RA << '\t' << center_DEC << '\t'
         << X_size << '\t' << Y_size <<'\t' << begin_Time << '\t' << end_Time << endl;
    pointing *pt = new pointing();
    pt->phi = center_RA * PI / 180;
    pt->theta = (90-center_DEC) * PI /180;
    int healpix_ID = HB.ang2pix(*pt);
    query_engine(sources[0], healpix_ID, center_RA, center_DEC, X_size, Y_size, begin_Time, end_Time);
    delete(pt);
    return 0;
}
int query_engine(string source, int hid, double center_RA, double center_DEC, double RA_size, double Dec_size, string begin_Time, string end_Time)
{
    stringstream ss;
    ss << "SELECT DISTINCT filepath, center_dec, center_ra, rotation FROM " << source << " WHERE hid = " << hid
       << " AND " << " obstime > '" << begin_Time << "' AND obstime < '" << end_Time <<"'" ;
    //cout << ss.str() << endl;
    int num = 0;
    try
    {
        openDBConnection();
        /* Create SQL statement */

        /* Create a non-transactional object. */
        nontransaction N(*C);

        /* Execute SQL query */
        result R(N.exec( ss.str().c_str() ));

        /* List down all the records */
        num = selectRows(R, center_RA, center_DEC, RA_size, Dec_size);

        //cout << "Operation done successfully" << endl;
        closeDBConnection();
    }
    catch (const std::exception &e)
    {
        cerr << e.what() << std::endl;
        return -1;
    }
    return num;
}


/* Command Line Interface */
int grammarAnalysis()
{
    string input;
    double center_RA, center_DEC, X_size, Y_size;
    string begin_Time, end_Time;
    cin >> input;
    transform(input.begin(), input.end(), input.begin(), ::tolower);
    while ( (input != string("\\q")))
    {
        if ( input == "select" )
        {
            cin >> input;
            transform(input.begin(), input.end(), input.begin(), ::tolower);
            if ( input == string("subimage") )
            {
                cin >> input;
                transform(input.begin(), input.end(), input.begin(), ::tolower);
                if (input != "from")
                {
                    cerr << "[FROM] expected" << endl;
                }
                cin >> input;
                string sources [100];
                int numOfSources = 0;
                size_t pos1 = -1, pos2 = input.find(',', 0);
                do
                {
                    sources[numOfSources++] = input.substr(pos1+1,pos2);
                    pos1 = pos2;
                    pos2 = input.find(',', pos1);
                }
                while (pos2 != string::npos);
                cin >> input;
                transform(input.begin(), input.end(), input.begin(), ::tolower);
                if (input != string("where"))
                {
                    cerr << "Error after source list" << endl;
                }
                else
                {
                    string selectionList;
                    cin >> input;
                    transform(input.begin(), input.end(), input.begin(), ::tolower);
                    while (input.at(input.length()-1) != ';')
                    {
                        selectionList.append(input);
                        cin >> input;
                        transform(input.begin(), input.end(), input.begin(), ::tolower);
                    }
                    selectionList.append(input, 0, input.length()-1);
                    //cout << "lists: " << selectionList << endl;
                    istringstream ss(selectionList);
                    string token;
                    string str1, str2;
                    while(getline(ss, token, ','))
                    {
                        //here we should have got like "center_ra=12.34"

                        cout << token  << endl;
                        str1 = token.substr(0, token.find('='));
                        str2 = token.substr(token.find('=')+1, string::npos);
                        transform(str1.begin(), str1.end(), str1.begin(), ::tolower);
                        transform(str2.begin(), str2.end(), str2.begin(), ::tolower);
                        if (str1 == string("center_ra"))
                        {
                            center_RA = atof(str2.c_str());
                        }
                        else if (str1 == string("center_dec"))
                        {
                            center_DEC = atof(str2.c_str());
                        }
                        else if (str1 == string("x_size"))
                        {
                            X_size = atof(str2.c_str());
                        }
                        else if (str1 == string("y_size"))
                        {
                            Y_size = atof(str2.c_str());
                        }
                        else if (str1 == string("begin_time"))
                        {
                            begin_Time = str2;
                        }
                        else if (str1 == string("end_time"))
                        {
                            end_Time = str2;
                        }
                        else
                        {
                            cerr << "Parse [SELECTION_LIST] error: '" << token << "'" << endl;
                        }
                    }
                    query_engine(sources,center_RA, center_DEC, X_size, Y_size, begin_Time, end_Time);
                }
            }
            else if (input == string("rawimage") )
            {

            }
            else if (input == string("reconstruct") )
            {

            }
            else
            {
                cerr << "Syntax error after SELECT" << endl;
            }
        }
        else
        {
            cerr << "Not beginning with [SELECT] " << input << endl;
        }

        cin >> input;
        transform(input.begin(), input.end(), input.begin(), ::tolower);
    }
    return 0;
}



/* check if two pixels are neighbors*/
bool haveCommonEdge(const unsigned long hid1, const unsigned long hid2) {
    string hid1_str = binary(hid1), hid2_str = binary(hid2);
    unsigned long rVal;
    int i;
    int pos1;
    string temp;

    pos1 = max(cfind(hid1_str,"00"), cfind(hid1_str, "10"));
    if (pos1 >= 0) {
        temp = hid1_str;
        for ( i = 0; i < pos1; i++) {
            temp.at(i) = '0';
        }
        for ( ; (unsigned)i < temp.length(); i+=2) {
            temp.at(i) = '0';
            temp.at(i+1) = '1';
        }
        rVal = strtoul(temp.c_str(), NULL, 2);
        if ((hid1 ^ rVal) == hid2) {
            return true;
        }
    }

    pos1 = max(cfind(hid1_str, "11"), cfind(hid1_str,"01"));
    if (pos1 >= 0) {
        temp = hid1_str;
        for ( i = 0; i < pos1; i++) {
            temp.at(i) = '0';
        }
        for ( ; (unsigned)i < temp.length(); i+=2) {
            temp.at(i) = '0';
            temp.at(i+1) = '1';
        }
        rVal = strtoul(temp.c_str(), NULL, 2);
        if ((hid1 ^ rVal) == hid2) {
            return true;
        }
    }

    pos1 = max(cfind(hid1_str, "11"), cfind(hid1_str,"10"));
    if (pos1 >= 0) {
        temp = hid1_str;
        for ( i = 0; i < pos1; i++) {
            temp.at(i) = '0';
        }
        for ( ; (unsigned)i < temp.length(); i+=2) {
            temp.at(i) = '1';
            temp.at(i+1) = '0';
        }
        rVal = strtoul(temp.c_str(), NULL, 2);
        if ((hid1 ^ rVal) == hid2) {
            return true;
        }
    }

    pos1 = max(cfind(hid1_str, "00"), cfind(hid1_str,"01"));
    if (pos1 >= 0) {
        temp = hid1_str;
        for ( i = 0; i < pos1; i++) {
            temp.at(i) = '0';
        }
        for ( ; (unsigned)i < temp.length(); i+=2) {
            temp.at(i) = '1';
            temp.at(i+1) = '0';
        }
        rVal = strtoul(temp.c_str(), NULL, 2);
        if ((hid1 ^ rVal) == hid2) {
            return true;
        }
    }

    return false;

}

/* convert coordinate system*/ 
void standardToRotatedCoordiantes(const double &x1, const double &y1, double &x2, double &y2, const double &rotation)
{
    rotatedToStandardCoordiantes(x1, y1, x2, y2, PI*2-rotation);
}

void rotatedToStandardCoordiantes(const double &x1, const double &y1, double &x2, double &y2, const double &rotation)
{
    x2 = x1*cos(rotation)-y1*sin(rotation);
    y2 = x1*sin(rotation)+y1*cos(rotation);
}


/* Database operations*/
static connection *C;

int openDBConnection()
{
    try {
      C = new connection ("dbname=fits_index user=postgres password=postgres \
      hostaddr=127.0.0.1 port=5432");
      //C->set_client_encoding("UTF8");
      if (C->is_open()) {
         //cout << "Opened database successfully: " << C->dbname() << endl;
      } else {
         cout << "Can't open database" << endl;
         return 1;
      }
    }
    catch (const std::exception &e){
        throw e;
    }
    return 0;
}
int closeDBConnection()
{
    try {
       C->disconnect ();
    }
    catch (const std::exception &e){
        throw e;
        return 1;
    }
    return 0;
}
int insertOne(string filepath, string filename, long hid, double ra, double dec, double rotation, string time)
{
    try{
        /* Create SQL statement */
        stringstream sql;
        sql << "INSERT INTO "<<DB_TABLE<<" (filepath, filename, hid, center_ra, center_dec, rotation, obstime)"
            << " VALUES ('"<< filepath << "','" << filename << "'," << hid << ","
            << ra << "," << dec << "," << rotation << ", '" << time << "');";
        string s = sql.str();
        /* Create a transactional object. */
        work W(*C);

        /* Execute SQL query */
        W.exec( sql.str().c_str() );
        W.commit();
        //cout << "Records created successfully" << endl;
    }
    catch (const std::exception &e){
        cerr << e.what() << std::endl;
        return 1;
    }
    return 0;
}





string binary( unsigned long n )
{
  string result;

  do result.push_back( '0' + (n & 1) );
  while (n >>= 1);

  reverse( result.begin(), result.end() );
  return result;
}

int cfind(string str, string sub)  //if not found, return -1
{
    int n = str.length()-2;

    for ( ; (n >= 0) && (str.substr(n, 2) != sub); n-=2 );
    return (n >= 0)? n:-1;
}

int randomQuery()
{
    double center_ra = (rand() % 25000 + 5000) / 100;
    double center_dec = (rand() % 2500 + 6500) / 100;
//    double x_size = (rand() % 16 + 5) / 10;
//    double y_size = (rand() % 16 + 5) / 10;
    double x_size = x_length * PERCENTAGE;
    double y_size = y_length * PERCENTAGE;
    pointing pt1;
    pt1.phi = center_ra* PI / 180;
    pt1.theta = (90-center_dec) * PI /180;
    return query_engine(DB_TABLE, HB.ang2pix(pt1), center_ra, center_dec, x_size, y_size, "2010-01-01", "2015-01-01");
}

int main()
{
    openDBConnection();
    for (NSIDE = 2; NSIDE <= 8; NSIDE++) {
        stringstream ss;
        ss << "fileindex_500g_0d125x0d0625_" << NSIDE;
        DB_TABLE = ss.str();
        ftw("/home/hong/disks/disk1/Mohe/test",fn,500);
    }
    closeDBConnection();
    return 0;
}
